// -*- C++ -*- 
//
// gsim_gen.cc
// Single event generator
//
// author  : Mitsuhiro YAMAGA
// created : Thu May 10 13:40:31 JST 2001
//
// $Id$
//
// Revision history
//
// $Log$
//

#include <iostream>
#include "math.h"
#include "CLHEP/Vector/ThreeVector.h"
#include "ccfortran/ccfortran.h"
#include "Ptype.h"
#include "GsimGeneral.h"


//////
#define  DEBUG  0


////////////////////////////////
// variables
static long         g_seeds[2]  = {0,0};
static int          g_seedindex = 0;


////////////////////////////////
// functions

// random
static inline double
randomDouble( double dmin,double dmax )
{
  // welcome
  //std::cout << "randomDouble" << std::endl;

  float r[1];
  GRNDM( r,1 );
  return( dmin + r[0]*(dmax - dmin) );
}


// random
static inline double
randomGauss( double mean,double sigma )
{
  // welcome
  //std::cout << "randomDouble" << std::endl;

  float r[1];
  RNORMX( r,1, grndm_ );
  return( mean + r[0]*sigma );
}


// random
static Hep3Vector
randomVector3( int mode,
               double x1min, double x1max,
               double x2min, double x2max,
               double x3min, double x3max )
{
  // welcome
  //std::cout << "randomVector3" << std::endl;

  // mode = 0 : X-Y-Z, uniform density
  // mode = 1 : R-Phi-Z, uniform density
  // mode = 2 : R-Theta-Phi
  // mode = 3 : R-CosTh-Phi
  
  float  rr[3];
  GRNDM( rr,3 );
  if ( mode == 0 ){
    return( Hep3Vector( x1min + rr[0]*(x1max - x1min),
                        x2min + rr[1]*(x2max - x2min),
                        x3min + rr[2]*(x3max - x3min)  ) );
  }
  else if ( mode == 1 ){ // r-phi-z, uniform in r-phi plane
    double ru = sqrt( rr[0] );
    double r = x1min + ru*(x1max - x1min);
    double p = x2min + rr[1]*(x2max - x2min);
    double z = x3min + rr[2]*(x3max - x3min);
    return( Hep3Vector( r*cos(p),r*sin(p),z ) );
  }
  else if ( mode == 2 ){ // r-theta-phi
    double r = x1min + rr[0]*(x1max - x1min);
    double t = x2min + rr[1]*(x2max - x2min);
    double p = x3min + rr[2]*(x3max - x3min);
    return( Hep3Vector( r*sin(t)*cos(p),r*sin(t)*sin(p),r*cos(t) ) );
  }
  else if ( mode == 3 ){ // r-cosTheta-phi
    double r = x1min + rr[0]*(x1max - x1min);
    double tmin = 1 - cos(x2min);
    double tmax = 1 - cos(x2max);
    double t = acos( 1 - ( tmin + rr[1]*(tmax - tmin) ) );
    double p = x3min + rr[2]*(x3max - x3min);
    return( Hep3Vector( r*sin(t)*cos(p),r*sin(t)*sin(p),r*cos(t) ) );
  }
  return( Hep3Vector( 0,0,0 ) );
  //
}


////////////////////////////////
// genParticle
void
GsimParticleGun( int pid, 
                 double pmin,  double pmax,
                 double pthmin,double pthmax, 
                 double pphmin,double pphmax,
                 double rmin,  double rmax,
                 double phmin, double phmax,
                 double zmin,  double zmax,
                 Hep3Vector& p, Hep3Vector& vtx  )
{
  // welcome
  //std::cout << "genParticle" << std::endl;

  // generate events
  vtx = randomVector3( 1, rmin,rmax, phmin,phmax, zmin,zmax );
  p   = randomVector3( 3, pmin, pmax,
                       deg2rad(pthmin),deg2rad(pthmax),
                       deg2rad(pphmin),deg2rad(pphmax) );
  
  //std::cout << vtx << std::endl;
  //std::cout << p   << std::endl;
  
}


////////////////////////////////
// Klong decay mode
void
GsimKlongDecayMode( int decayMode )
{
  float  bratio[6] = { 100,0,0,0,0,0 };
  int    mode[6]   = { 0,0,0,0,0,0 };
  enum { gm=1,ep,em,nu,mp,mm,p0,pp,pm,KL };
  
  if ( decayMode == 1 )
    mode[0] = pm + ep*100 + nu*10000;
  if ( decayMode == 2 )
    mode[0] = pp + em*100 + nu*10000;
  if ( decayMode == 3 )
    mode[0] = pm + mp*100 + nu*10000;
  if ( decayMode == 4 )
    mode[0] = pp + mm*100 + nu*10000;
  if ( decayMode == 5 )
    mode[0] = p0 + p0*100 + p0*10000;
  if ( decayMode == 6 )
    mode[0] = pp + pm*100 + p0*10000;

  if ( decayMode == 7  ){ // KL -> pi0 pi0
    mode[0] = p0 + p0*100;
  }
  if ( decayMode == 8  ){ // KL -> pi+ pi-
    mode[0] = pp + pm*100;
  }
  if ( decayMode == 9  ){ // KL -> gamma gamma
    mode[0] = gm + gm*100;
  }
  if ( decayMode == 10 ){ // KL -> pi0 nu nubar
    mode[0] = p0 + nu*100 + nu*10000;
  }

  if ( decayMode == 99 ){
    bratio[0] = 0;
    mode[0] = 0; // no decay
  }

  if ( decayMode )
    GSDK( KL,bratio,mode );

}


// Momentum distribution at the Target+6cm.
static double
PklDist_Target( double p )
{
  // parameters from H.Watanabe, Mon, 23 Jul 2001 19:34:16
  const static double  norm        =  1.0;
  const static double  mean        =  1.0317;
  const static double  sigma       =  0.6706;
  const static double  asym_offset = -0.48365;
  const static double  asym_slope  =  0.013031;
  
  double  sig = sigma * (1.0 - (asym_offset + p * asym_slope)*(p - mean) );
  double  f   = norm * exp( -0.5*((p-mean)*(p-mean))/(sig*sig) );

  return( f ); // yield
}


// Momentum distribution at the exit of collimator6
static double
PklDist_Col6_Open( double p )
{
  // MC data from M.Yamaga,
  // fitting function by H.Watanabe,
  // fitting parameter from M.Yamaga, Thu Sep 27 01:08:29 JST 2001
  const static double  norm        =  1.0;
  const static double  mean        =  1.9319;
  const static double  sigma       =  1.6679;
  const static double  asym_offset = -0.34098;
  const static double  asym_slope  =  2.42863e-2;
  
  double  sig = sigma * (1.0 - (asym_offset + p * asym_slope)*(p - mean) );
  double  f   = norm * exp( -0.5*((p-mean)*(p-mean))/(sig*sig) );

  return( f ); // yield
}


// Momentum distribution at the exit of collimator6
static double
PklDist_Col6_Pb( double p )
{
  // MC data from S.Inoue,
  // fitting function by H.Watanabe,
  // fitting parameter from M.Yamaga, Thu Sep 27 01:08:29 JST 2001
  const static double  norm        =  1.0;
  const static double  mean        =  2.0270;
  const static double  sigma       =  1.7385;
  const static double  asym_offset = -0.31790;
  const static double  asym_slope  =  2.31968e-2;
  
  double  sig = sigma * (1.0 - (asym_offset + p * asym_slope)*(p - mean) );
  double  f   = norm * exp( -0.5*((p-mean)*(p-mean))/(sig*sig) );

  return( f ); // yield
}


// Momentum distribution on the E391a beamline, 15m - 20m in Z position.
static double
PklDist_15m_20m( double p )
{
  // parameters from H.Watanabe, Mon, 23 Jul 2001 19:34:16
  const static double  norm        =  1.0;
  const static double  mean        =  0.93132;
  const static double  sigma       =  0.51036;
  const static double  asym_offset = -0.54786;
  const static double  asym_slope  =  0.013947;
  
  double  sig = sigma * (1.0 - (asym_offset + p * asym_slope)*(p - mean) );
  double  f   = norm * exp( -0.5*((p-mean)*(p-mean))/(sig*sig) );

  return( f ); // yield
}


// Density distribution on the E391a beamline at collimator6
static double
Density_Col6( double r )
{
  // Data from S.Inoue.
  // parameters from M.Yamaga, Fri Sep 28 13:19:17 JST 2001
  // Function : Fermi dist. + 1st order polynomial with p0=0.
  const static double  norm  =  1.0;
  const static double  x0    =  1.9095;
  const static double  slope =  0.18337;
  const static double  p0    =  0;
  const static double  p1    = -0.10679e-01;
  
  double  f = 1.0 / (1 + exp((r-x0)/slope)) + (p0 + p1*r);

  return( f ); // yield
}


////////////////////////////////
// Generate Klong
void
GsimGenKlongCol6( double z, int absorber, Hep3Vector& p, Hep3Vector& vtx )
{
  // normalization factor at position z(detector coordinate)
  const static double  ctau = 1551; // [cm] 
  const static double  mkl  = ptype[10].mass; // KL mass
  const static double  m_ct = mkl / ctau; 
  const static double  lbeamline = 987.5; //[cm]
  const static double  col6_z0 = TargetZ + lbeamline;

  double  bz = z - col6_z0; // bz0 = exit point of collimator6
  if ( bz < 0 )
    bz = 0;

  // vertex distribution
  // Data from S.Inoue
  // Parameter from M.Yamaga, Fri Sep 28 15:22:34 JST 2001

  // R of vertex at z0
  const static double  rmax = 5.24/2;
  double  rr,prob,yield;
  //static int  ct=0;
  do{
    rr    = sqrt( randomDouble( 0,rmax*rmax ) );
    yield = Density_Col6( rr );
    prob  = randomDouble( 0,1 );
    //std::cout << "rr=" << rr 
    //           << " prob=" << prob << " yield=" << yield 
    //          << " ct=" << ++ct
    //          << endl;
  } while( prob > yield );

  vtx = randomVector3( 1, rr, rr, 0, 2*M_PI, col6_z0, col6_z0 );// r-phi-z
  
  // Momentum direction at z0
  // Data from S.Inoue
  // Parameter from M.Yamaga, Fri Sep 28 15:22:34 JST 2001
  
  // Polar angle of momentum : uniform
  double  thmax =  0.0004 + (0.0035-0.0004)/3*rr;
  double  thmin = -0.0006 + (0.0033+0.0006)/3*rr;
  if ( thmin < 0 )
    thmin = 0;
  double  theta = randomDouble( thmin,thmax );

  // Residual of Azimuth angle of momentum : double Gaussian
  const static double  gnorm1  = 412.4;
  const static double  gmean1  = 0.6628e-03;
  const static double  gsigma1 = 0.1002;
  const static double  gnorm2  = 60.71;
  const static double  gmean2  = 0.2593e-03;
  const static double  gsigma2 = 0.2621;
  const static double  garea1  = gnorm1*gsigma1;
  const static double  garea2  = gnorm2*gsigma2;
  const static double  gfrac1  = garea1/(garea1+garea2);
  double  dphi;
  if ( randomDouble(0,1) < gfrac1 )
    dphi = randomGauss( gmean1,gsigma1 );
  else
    dphi = randomGauss( gmean2,gsigma2 );

  // Momentum magnitude distribution
  const static double  MinP = 0.01;
  const static double  MaxP = 11;
  double  ProbMax = exp( -m_ct*bz/MaxP ); // exp(-t/tau)=exp((M*Z)/(c*tau*P))
  double  pp,norm;
  //static int  ct=0;
  do {
    pp    = randomDouble( MinP, MaxP );
    norm  = exp( -m_ct*bz/pp ); // exp(-t/tau)=exp((M*Z)/(c*tau*P))
    prob  = randomDouble( 0, ProbMax );
    if ( absorber == 0 )
      yield = PklDist_Col6_Open( pp ) * norm;
    else if ( absorber == 1 )
      yield = PklDist_Col6_Pb( pp ) * norm;
    else
      yield=0;
    //std::cout << "p=" << pp << " norm=" << norm << " PMax=" << ProbMax 
    //          << " prob=" << prob << " yield=" << yield 
    //          << " ct=" << ct++
    //          << endl;
  } while ( prob > yield );

  p = Hep3Vector( 0,0,pp );

  // All parameter decided

  // set momentum direction at position z
  p.setTheta( theta );
  p.setPhi( vtx.phi() + dphi );

  // set vertex position at position z
  //vtx += p/p.z()*bz;
  vtx += p * ( bz / p.z() );

  if ( DEBUG )
    std::cout << "GsimGenKlongCol6 : v" << vtx << " :p" << p << std::endl;
  //
}


////////////////////////////////
// Generate Klong
void
GsimGenKlongTarget( double z, Hep3Vector& p, Hep3Vector& vtx )
{
  // normalization factor at position Z
  const static double  ctau = 1551; // [cm] 
  const static double  mkl  = ptype[10].mass; // KL mass
  const static double  m_ct = mkl / ctau; 

  double  bz = z - TargetZ; // Beamline coordinate ( target = z0 )
  if ( bz < 0 )
    bz = 0;

  const static double  MinP = 0.01;
  const static double  MaxP = 11;
  double  ProbMax = exp( -m_ct*bz/MaxP ); // exp(-t/tau)=exp((M*Z)/(c*tau*P))
 
  double  pp,prob,yield,norm;
  //static int  ct=0;
  do {
    pp    = randomDouble( MinP, MaxP );
    norm  = exp( -m_ct*bz/pp ); // exp(-t/tau)=exp((M*Z)/(c*tau*P))
    prob  = randomDouble( 0, ProbMax );
    yield = PklDist_Target( pp ) * norm;
    //std::cout << "p=" << pp << " norm=" << norm << " PMax=" << ProbMax 
    //          << " prob=" << prob << " yield=" << yield 
    //          << " ct=" << ct++
    //          << endl;
  } while ( prob > yield );
  
  p   = Hep3Vector( 0,0,pp );
  vtx = randomVector3( 1, 0, 2, 0, 2*M_PI, bz+TargetZ, bz+TargetZ );// r-phi-z

  if ( DEBUG )
    std::cout << "GsimGenKlongTarget : v" << vtx << " :p" << p << std::endl;
  //
}


